#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "editrec.h"

#define MARGIN 3
#define RMARGIN 8
#define ETEXT "etext"

EditRec *currentrec = NULL;

/* internal prototypes */
void CalcLineBreaks(EditRec *);
boolean GetNextWord(BigString, char *, int);
int FindLastChar(HugeString, int);
boolean ProcessVertKey(EditRec *, char ch);
boolean ProcessBackspace(EditRec *);
boolean ProcessTypingChar(EditRec *, char ch);

void UpdateEditRec(EditRec *rec)
{
	CalcLineBreaks(rec);
	DrawEditRec(rec);
}

void SetCurrentEditRec(EditRec *er)
{
	EditRec *oldcurrent = currentrec;
	
	if (er == currentrec)
		return;
		
	currentrec = er;
	
	DrawEditRec(er);
	if (oldcurrent != NULL)
		DrawEditRec(oldcurrent);
}

EditRec *GetCurrentEditRec()
{
	return currentrec;
}

EditRec *NewEditRec(char *text, int maxtlen, int maxtlines, 
		FontSizeType fsize, char *windowname, char *dragareaname, 
		boolean (*keyfilter)(char *), void (*notification)(void))
{
	EditRec *newrec = (EditRec *) malloc(sizeof(EditRec));
	
	if (newrec == NULL)
		GSelog(-1, "Out of memory for EditRec.\n");
		/*return NULL;*/
		
	newrec->text = text;
	newrec->maxtlen = maxtlen;
	newrec->maxtlines = maxtlines;
	newrec->fsize = fsize;
	newrec->keyfilter = keyfilter;
	newrec->notification = notification;
	
	newrec->inspt = 0;
	
	newrec->linebreaks = (int *) malloc(sizeof(int)*maxtlines);
	if (newrec->linebreaks == NULL)
		GSelog(-1, "Out of memory for EditRec.\n");		

	newrec->windowname = malloc(strlen(windowname)+1);
	if (newrec->windowname == NULL)
		GSelog(-1, "Out of memory for EditRec.\n");		

	newrec->dragareaname = malloc(strlen(dragareaname)+1);
	if (newrec->dragareaname == NULL)
		GSelog(-1, "Out of memory for EditRec.\n");		
	
	strcpy(newrec->windowname, windowname);	
	strcpy(newrec->dragareaname, dragareaname);	

	CalcLineBreaks(newrec);
	
	return newrec;
}

void CalcLineBreaks(EditRec *rec)
{
	int textpos = 0, wordsplaced = 0;
	int wordwidth, linewidth = 0, maxlinewidth, wd;
	HugeString nextword;
	GSColor col;
	
	if (rec == NULL)
		return;
		
	rec->lastline = 0;
	
	GSgetbuttonspecs(rec->windowname, rec->dragareaname, 
								NULL, NULL, &wd, NULL, NULL);
	
	maxlinewidth = wd - MARGIN - RMARGIN;
	
	while (GetNextWord(nextword, rec->text, textpos))	{

		wordwidth = GStextwidth(nextword);

		if (wordwidth + linewidth >= maxlinewidth)	{ /* won't fit */
			
			if (*nextword == ' ')
				textpos++;

			if (wordsplaced == 0)	{
				textpos += FindLastChar(nextword, maxlinewidth);
				}
				
			rec->linebreaks[rec->lastline] = textpos;
			linewidth = 0;
			wordsplaced = 0;
			
			if (rec->lastline < rec->maxtlines-1 )
				rec->lastline++;
			else
				return;
			}
		else	{
			linewidth += wordwidth;
			textpos += strlen(nextword);
			wordsplaced++;
			}			
		}
	
	if (rec->lastline < rec->maxtlines)	/* this should always be the case */
		rec->linebreaks[rec->lastline] = textpos;
		
}

boolean GetNextWord(HugeString nextword, char *text, int textpos)
{
	int i = 0;
	
	if (strlen(text) <= textpos)
		return FALSE;
		
	nextword[0] = text[textpos];
	
	for(i=1; (text[i+textpos] != ' ') && (text[i+textpos] != '\0')
						&& (i < HUGESTRINGLEN); i++)
		nextword[i] = text[i+textpos];
	
	nextword[i] = '\0';
	
	return TRUE;
}

int FindLastChar(HugeString nextword, int maxlinewidth)
{
	int last = strlen(nextword);
	
	do	{
		nextword[--last] = '\0';
		if (last <= 1)
			break;
		} while (GStextwidth(nextword) > maxlinewidth);
		
	return (last);
}

void DrawEditRec(EditRec *rec)
{
	int x, y, wd, ht, lineht = GStextheight(rec->text), width = 0, dx, dy, ch_width;
	int curline, curtextpos = 0, textsegend;
	char tempch;
	boolean active = (rec == currentrec);
	
	GSsetcurrentnamedwindow(rec->windowname);
	GSsetactivenamedwindow(rec->windowname);
	GSsetlinesize(1);
	GSgetwindowspecs(rec->windowname, &dx, &dy, NULL, NULL, NULL);
	GSgetbuttonspecs(rec->windowname, rec->dragareaname, &x, &y, &wd, &ht, NULL);
	
	GSsetcolor(GSconvertcolor("white"));
	GSwfillrect(x, y, wd, ht);
	
	GSsetcolor(GSconvertcolor("black"));
	GSwdrawline(x, y, x + wd, y);
	GSwdrawline(x, y, x, y + ht);
	GSsetcolor(GSconvertcolor(ETEXT));
	GSwdrawline(x + wd, y, x + wd, y + ht);
	GSwdrawline(x, y + ht, x + wd, y + ht);
	GSsetfont(rec->fsize);
	
	x += MARGIN;
	y += 2*MARGIN;
	ht -= 2*MARGIN;
	wd -= MARGIN + RMARGIN;
	
	GSsetcolor(active? GSconvertcolor("black") : GSconvertcolor(ETEXT));
	
	for	(curline = 0; curline <= rec->lastline; curline++)	{
		y += lineht;
		if ((rec->inspt >= curtextpos) && (active) &&
					((curline == rec->lastline) || (rec->inspt < rec->linebreaks[curline]))) {
						
			textsegend = MIN(rec->inspt, rec->linebreaks[curline]);
			if (textsegend > curtextpos)	{ 
				tempch = rec->text[textsegend];
				rec->text[textsegend] = '\0';
				GSwwritemsg(x, y, &(rec->text[curtextpos]));
				width = GStextwidth(&(rec->text[curtextpos]));
				rec->text[textsegend] = tempch;
				
				curtextpos = textsegend;
				}
			if (rec->inspt <= rec->linebreaks[curline]) {		

#if 0
	/* old way */
				GSPxormode(TRUE);
				if (rec->text[textsegend] == '\0')
					GSwwritemsg(x+width, y, " ");
				else {
					tempch = rec->text[textsegend+1];
					rec->text[textsegend+1] = '\0';
					GSwwritemsg(x+width, y, &(rec->text[textsegend]));
					width += GStextwidth(&(rec->text[textsegend]));
					rec->text[textsegend+1] = tempch;
					}
				GSPxormode(FALSE);
#else				
				if (rec->text[textsegend] != '\0') {
					tempch = rec->text[textsegend+1];
					rec->text[textsegend+1] = '\0';
					GSwwritemsg(x+width, y, &(rec->text[textsegend]));
					ch_width = GStextwidth(&(rec->text[textsegend]));
					rec->text[textsegend+1] = tempch;
					}
				else {
					GSwwritemsg(x+width, y, " ");
					ch_width = GStextwidth(" ");
					}
/*				GSPxormode(TRUE);*/
				GSwdrawline(x+width+1, y, x+width+ch_width, y);
/*				GSPxormode(FALSE);*/
				width += ch_width;
#endif
				curtextpos = textsegend+1;
				if (curtextpos < rec->linebreaks[curline])	{
					tempch = rec->text[rec->linebreaks[curline]];
					rec->text[rec->linebreaks[curline]] = '\0';
					GSwwritemsg(x+width, y, &(rec->text[curtextpos]));
					rec->text[rec->linebreaks[curline]] = tempch;
					}
				}
			}
	else	{
		tempch = rec->text[rec->linebreaks[curline]];
		rec->text[rec->linebreaks[curline]] = '\0';
		GSwwritemsg(x, y, &(rec->text[curtextpos]));
		rec->text[rec->linebreaks[curline]] = tempch;
		}
	curtextpos = rec->linebreaks[curline];
	y += lineht / 2;
	}
}

void DisposEditRec(EditRec *rec)
{
	free(rec->linebreaks);
	free(rec->windowname);
	free(rec->dragareaname);
	if (rec == currentrec)
		currentrec = NULL;
	free(rec);
}

boolean ProcessChar(char chr)
{
	char ch = chr;
	boolean refresh = TRUE;
	
	if (currentrec == NULL)
		return FALSE;
	
	switch(ch)	{
	
		case KeyUp:
		case KeyDown:
			refresh = ProcessVertKey(currentrec, ch);
			break;
			
		case KeyLeft:
			if (currentrec->inspt > 0)	{
				refresh = TRUE;
				currentrec->inspt--;
				}
			else
				refresh = FALSE;
			break;
		
		case KeyRight:
			if (currentrec->inspt < strlen(currentrec->text)) {
				refresh = TRUE;
				currentrec->inspt++;
				}
			else
				refresh = FALSE;
			break;
			
		case '\b':
			refresh = ProcessBackspace(currentrec);
			break;
			
		default:
			refresh = ProcessTypingChar(currentrec, ch);
			break;
		}
		
	if (refresh)	{
		CalcLineBreaks(currentrec);
		DrawEditRec(currentrec);
		if (currentrec->notification != NULL)
			currentrec->notification();
		}
	else
		GSbell();
		
	return TRUE;
}

boolean ProcessVertKey(EditRec *rec, char ch)
{
	int whichline, offset, newlinestart;
	
	for (whichline = 0; (rec->inspt >= rec->linebreaks[whichline])
							&& (whichline < rec->lastline); whichline++);

	if (whichline > 0)
		offset = rec->inspt - rec->linebreaks[whichline-1];
	else
		offset = rec->inspt;
		
	if (ch == KeyUp) {
		if (whichline > 0)	{
			if (whichline == 1)
				newlinestart = 0;
			else
				newlinestart = rec->linebreaks[whichline-2];
			rec->inspt = MIN(newlinestart + offset, rec->linebreaks[whichline-1]-1);
			}
		else
			return FALSE;
		}
	else if (ch == KeyDown) {
		if (whichline < rec->lastline)	{
			newlinestart = rec->linebreaks[whichline];
			rec->inspt = MIN(offset + newlinestart, rec->linebreaks[whichline+1]-1);
			}
		else
			return FALSE;
		}
	
	return TRUE;
}

boolean ProcessBackspace(EditRec *rec)
{
	int chindex, len;
	
	if (rec->inspt <= 0)
		return FALSE;
	
	chindex = rec->inspt;
	len = strlen(rec->text);
	
	rec->inspt--;

	for (chindex = rec->inspt; (chindex < len); chindex++)
		rec->text[chindex] = rec->text[chindex+1];
		
	return TRUE;
}

boolean ProcessTypingChar(EditRec *rec, char chr)
{
	int chindex, len;
	char ch = chr;

	if ((chr < ' ') || (chr > '~'))
		return TRUE;
	
	if (rec->keyfilter != NULL)	{
		if (!currentrec->keyfilter(&ch))
			return FALSE;
		}
	
	len = strlen(rec->text);
	
	if (len >= rec->maxtlen)
		return FALSE;
		
	for (chindex = len; (chindex >= rec->inspt); chindex--)
		rec->text[chindex+1] = rec->text[chindex];
		
	rec->text[rec->inspt] = ch;
	rec->inspt++;
	
	return TRUE;
}

void SetEditRecText(EditRec *rec, char *tbuf, int maxlen)
{
	rec->text = tbuf;
	rec->maxtlen = maxlen;
	UpdateEditRec(rec);
}

/* Common key filters */

boolean boolean_filter(char *ch)
{
	return ((*ch == '0') || (*ch == '1'));
}

boolean numeric_filter(char *ch)
{
/*	return isdigit(ch); UNIX doesn't like this - why? */
	return ((*ch >= '0') && (*ch <= '9'));
}

boolean alphabetic_filter(char *ch)
{
	return (((*ch>='a')&&(*ch<='z'))
				|| ((*ch>='A')&&(*ch<='Z')));
}

boolean alphanumeric_filter(char *ch)
{
/*	return isalnum(ch);*/
	return (numeric_filter(ch) || alphabetic_filter(ch));
}

boolean c_identifier_filter(char *ch)
{
	return (alphanumeric_filter(ch) || (*ch == '_'));
}
